from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import Qt

from .docks import DockBase
       
class AreaSplitterHandle(QtWidgets.QSplitterHandle):

    def mouseDoubleClickEvent(self, event):
        self.parent().distributeWidgets()        

class DockBox(QtWidgets.QSplitter, DockBase):
    SplitArea = 0
    PinArea = 1
    ScrollArea = 2

    def __init__(self, orientation):
        QtWidgets.QSplitter.__init__(self, orientation)
        
        pal = self.palette()     
        if orientation == Qt.Horizontal:
            pal.setColor(QtGui.QPalette.Background, QtGui.QColor(192,224,192))
        elif orientation == Qt.Vertical:
            pal.setColor(QtGui.QPalette.Background, QtGui.QColor(192,192,224))
        self.setPalette(pal)          
        
        self.pinarea = PinnedBox(orientation)
        self.scrollarea = ScrollBox(orientation)        
        
        super().addWidget(self.pinarea)
        super().addWidget(self.scrollarea)        
        
        self.collapsed = False
        self.scrollarea.hide()
        
    def createHandle(self):
        return AreaSplitterHandle(self.orientation(), self)         

    def addWidget(self, widget, area=PinArea, title=None):
        if area == DockBox.PinArea:
            self.pinarea.addWidget(widget)
            
        elif area == DockBox.ScrollArea:
            if self.scrollarea.count() == 0:
                self.scrollarea.show()
                self.setSizes([4,1], DockBox.SplitArea)
            self.scrollarea.addWidget(widget)
            
    def count(self, area=PinArea):
        if area == DockBox.SplitArea:
            return super().count()
            
        elif area == DockBox.PinArea:
            return self.pinarea.count()
            
        elif area == DockBox.ScrollArea:
            return self.scrollarea.count()          
        

    def moveToOtherArea(self, widget, area=None):
        if area is None:
            area_widget = widget.parent()
            while not isinstance(area_widget, (PinnedBox, ScrollBox)):
                area_widget = area_widget.parent()
            if isinstance(area_widget, PinnedBox):
                area = DockBox.ScrollArea
            elif isinstance(area_widget, ScrollBox):
                area = DockBox.PinArea
            
        if area == DockBox.ScrollArea:            
            self.addWidget(widget, area=DockBox.ScrollArea)
            
        elif area == DockBox.PinArea:
            self.scrollarea.removeWidget(widget)
            self.addWidget(widget, area=DockBox.PinArea)

    def distributeWidgets(self):
        sizes = self.pinarea.sizes()
        newsize = sum(sizes) // self.pinarea.count()
        sizes = [newsize] * self.pinarea.count()
        self.pinarea.setSizes(sizes)
        
    def setSizes(self, sizes, area=PinArea):
        if area == DockBox.SplitArea:
            super().setSizes(sizes)
            
        elif area == DockBox.PinArea:
            self.pinarea.setSizes(sizes)
            
        elif area == DockBox.ScrollArea:
            self.scrollarea.growbox.setSizes(sizes)    

    def sizes(self, area=PinArea):
        if area == DockBox.SplitArea:
            return super().sizes()

        elif area == DockBox.PinArea:
            return self.pinarea.sizes()

        elif area == DockBox.ScrollArea:
            return self.scrollarea.growbox.sizes()
    
    def currentIndex(self):
        return self.pinarea.currentIndex()
        
    def widget(self, index, area=PinArea):
        if area == DockBox.PinArea:
            return self.pinarea.widget(index)
            
        elif area == DockBox.ScrollArea:
            return self.scrollarea.growbox.widget(index)
            
    def show(self):        
        self.pinarea.show()
        
        if self.scrollarea.count() > 0:
            self.scrollarea.show()            
        else:
            self.scrollarea.hide()
            
        self.scrollarea.growbox.show()
        
        for pos in range(self.count(DockBox.PinArea)):
            widget = self.widget(pos, DockBox.PinArea)
            widget.show()
            
        for pos in range(self.count(DockBox.ScrollArea)):
            widget = self.widget(pos, DockBox.ScrollArea)
            widget.show()                           
            
        super().show()
        
    def setCurrentIndex(self, index):
        self.pinarea.setCurrentIndex(index)
    

class DockVBox(DockBox):

    def __init__(self):
        super().__init__(Qt.Vertical)
        
        
class DockHBox(DockBox):

    def __init__(self):
        super().__init__(Qt.Horizontal)   


class PinnedSplitterHandle(QtWidgets.QSplitterHandle):

    def mouseDoubleClickEvent(self, event):
        self.parent().distributeWidgets(self)
        

class PinnedBox(QtWidgets.QSplitter):

    def __init__(self, orientation):
        QtWidgets.QSplitter.__init__(self, orientation)

        pal = self.palette()     
        pal.setColor(QtGui.QPalette.Background, QtGui.QColor(192,192,192))
        self.setPalette(pal)
        
        self.atLeastOnePanel = True
        
    def collapse(self, widget):
        if self.atLeastOnePanel:        
            not_collapsed_widgets = [self.widget(pos) for pos in range(self.count()) if not self.widget(pos).collapsed]
            
            if not widget.collapsed and len(not_collapsed_widgets) == 1:
                widget.topleftbtn.setChecked(False)
                return        
            
        widget.collapseVertical()  
        
    def moveToOtherArea(self, widget):
        self.parent().moveToOtherArea(widget, DockBox.ScrollArea)
        
    def createHandle(self):
        return PinnedSplitterHandle(self.orientation(), self)   

    def distributeWidgets(self, handle):
        pos = self.indexOf(handle)
        
        sizes = self.sizes()
                        
        newsize = sum(sizes[pos-1:pos+1]) // 2
        sizes[pos-1] = newsize
        sizes[pos] = newsize        
        self.setSizes(sizes)           
        
        
class SplitterHandle(QtWidgets.QWidget):

    def __init__(self, orientation, space=5):
        super().__init__()
        self._orientation = orientation

        if orientation == Qt.Vertical:
            self.setCursor(QtGui.QCursor(QtCore.Qt.SplitVCursor))
            self.setFixedHeight(space)
            
        elif orientation == Qt.Horizontal:
            self.setCursor(QtGui.QCursor(QtCore.Qt.SplitHCursor))
            self.setFixedWidth(space)
            
    def orientation(self):
        return self._orientation
        
    def mouseDoubleClickEvent(self, event):
        self.parent().distributeWidgets(self)         
            
    def mouseMoveEvent(self, event):
        pos = self.parent().indexOf(self) - 1
            
        if self.orientation() == Qt.Vertical:
            delta = event.pos().y()
            self.parent().changeWidgetSize(pos, delta, delta=True)
            
        elif self.orientation() == Qt.Horizontal:
            delta = event.pos().x()
            self.parent().changeWidgetSize(pos, delta, delta=True)
        
        
class SplitBox(QtWidgets.QWidget):        

    def __init__(self, orientation):
        super().__init__()        
        self._orientation = orientation
        self.widgets = []
        self.handles = [None]
        self.space = 5
        
        if orientation == Qt.Vertical:
            layout = QtWidgets.QVBoxLayout()
            self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
            
        elif orientation == Qt.Horizontal:
            layout = QtWidgets.QHBoxLayout()                    
            self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Preferred)

        layout.setContentsMargins(0,0,0,0)
        layout.setSpacing(0)            
        self.setLayout(layout)
        self.atLeastOnePanel = False
        
    def orientation(self):
        return self._orientation
        
    def addWidget(self, widget):
        if self.orientation() == Qt.Vertical:
            sizes = self.sizes() + [widget.height()]
            
        elif self.orientation() == Qt.Horizontal:
            sizes = self.sizes() + [widget.width()]        
            
        self.widgets.append(widget)    
        self.layout().addWidget(widget)
        handle = SplitterHandle(self.orientation(), self.space)
        self.handles.append(handle)
        self.layout().addWidget(handle)
        self.layout().setStretch(self.layout().count()-1, self.space)
        
        self.setSizes(sizes)
        
    def insertWidget(self, pos, widget):
        if self.orientation() == Qt.Vertical:
            prefered_height = widget.height()
            sizes = self.sizes()
            sizes.insert(pos, prefered_height)
            
        elif self.orientation() == Qt.Horizontal:  
            prefered_width = widget.width()
            sizes = self.sizes()
            sizes.insert(pos, prefered_width)
            
        self.widgets.insert(pos, widget)
        self.layout().insertWidget(pos*2, widget)  
        handle = SplitterHandle(self.orientation(), self.space)
        self.handles.insert(pos+1, handle)
        self.layout().insertWidget(pos*2+1, handle)
        self.layout().setStretch(pos*2+1, self.space)
        
        self.setSizes(sizes)             

    def resizeWidget(self, widget, size):
        if isinstance(widget, int):
            pos = widget
        else:
            pos = self.indexOf(widget)
            
        sizes = self.sizes()                             
        sizes[pos] = size        
        self.setSizes(sizes)       

    def removeWidget(self, widget):
        if isinstance(widget, int):
            pos = widget
        else:
            pos = self.indexOf(widget)
            
        sizes = self.sizes()
        sizes.pop(pos)            
            
        widget = self.widgets.pop(pos)
        self.layout().removeWidget(widget)
        handle = self.handles.pop(pos+1)        
        self.layout().removeWidget(handle)        
        widget.setParent(None)
        handle.setParent(None)
        
        self.setSizes(sizes)
            
        return widget

    def count(self):
        return len(self.widgets)
        
    def widget(self, index):
        return self.widgets[index]
        
    def indexOf(self, widget):
        try:
            return self.widgets.index(widget)
        except ValueError:
            pass
            
        try:
            return self.handles.index(widget)
        except ValueError:
            pass            
            
        return None

    def distributeWidgets(self, handle):
        pos = self.indexOf(handle)
        sizes = self.sizes()
        
        cmo = self.count()
        
        if pos == cmo:
            sizes = [sum(sizes[:cmo]) // cmo] * cmo
        else:                    
            newsize = sum(sizes[pos-1:pos+1]) // 2
            sizes[pos-1] = newsize
            sizes[pos] = newsize   
            
        self.setSizes(sizes) 
        
    def collapse(self, widget):
        if self.atLeastOnePanel:        
            not_collapsed_widgets = [self.widget(pos) for pos in range(self.count()) if not self.widget(pos).collapsed]
            
            if not widget.collapsed and len(not_collapsed_widgets) == 1:
                widget.topleftbtn.setChecked(False)
                return
                
        widget.collapseVertical()

    def sizes(self):
        if self.orientation() == Qt.Vertical:
            return [widget.height() for widget in self.widgets]
        elif self.orientation() == Qt.Horizontal:
            return [widget.width() for widget in self.widgets]
        #return self.stretches()
            
    def stretches(self):
        return [self.layout().stretch(index) for index in range(self.count()*2)]
        
    def setSizes(self, sizes):
        total_size = sum(sizes) + self.space * self.count()
        # print(f'Sizes:    Stretches: {self.stretches()}')
        # print(f'    Old: {self.sizes()}')
        # print(f'    New: {sizes}   Total: {total_size}')
        
        if self.orientation() == Qt.Vertical:
            self.setFixedHeight(total_size)
        elif self.orientation() == Qt.Horizontal:
            self.setFixedWidth(total_size)            
        for (i, stretch) in zip(range(self.count()), sizes):            
            self.layout().setStretch(i*2, stretch)
            
        # QtWidgets.QApplication.instance().processEvents()            
        # print(f'    Eff: {self.sizes()}')
        # QtWidgets.QApplication.instance().processEvents()
        
    def changeWidgetSize(self, pos_or_widget, size, delta=False, fixed=False):
        """
        Set the widget at pos to size.
        Adapt the size of the splitbox so no other widgets resizes
        """
        sizes = self.sizes()
        
        if isinstance(pos_or_widget, int):
            pos = pos_or_widget
            widget = self.widget(pos)
        else:
            pos = self.indexOf(pos_or_widget)
            widget = pos_or_widget      
            
        if self.orientation() == Qt.Vertical:                     
            widget_height = sizes[pos]
            if delta: size = widget_height + size
            new_widget_height = max(size, widget.minimumHeight()) 
            sizes[pos] = new_widget_height                       
            if fixed:
                widget.setFixedHeight(size)
            
        elif self.orientation() == Qt.Horizontal:                 
            widget_width = sizes[pos]            
            if delta: size = widget_width + size
            new_widget_width = max(size, widget.minimumWidth())            
            sizes[pos] = new_widget_width                       
            if fixed:
                widget.setFixedWidth(size)        

        self.setSizes(sizes)  
        
        

class ScrollBox(QtWidgets.QScrollArea):

    def __init__(self, orientation):
        QtWidgets.QScrollArea.__init__(self)
        
        self.setWidgetResizable(True)
        
        if orientation == Qt.Vertical:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
            
        elif orientation == Qt.Horizontal:
            self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        
        pal = self.palette()     
        pal.setColor(QtGui.QPalette.Background, QtGui.QColor(160,200,224))
        self.setPalette(pal)           
        
        self.growbox = SplitBox(orientation)
        self.setWidget(self.growbox)        

    def addWidget(self, widget):
        self.growbox.addWidget(widget)
        
    def removeWidget(self, widget):
        self.growbox.removeWidget(widget)
            
    def count(self):
        return self.growbox.count()
    
        



import pathlib
import random
import logging
from collections import OrderedDict

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QKeySequence, QIcon, QPixmap, QCursor
from qtpy.QtWidgets import (QApplication, QWidget, QAction, QMainWindow, QDockWidget, QToolButton,
    QMdiSubWindow, QDockWidget, QShortcut, QPushButton, QMenu, QRadioButton, QButtonGroup)

from .. import gui, config
from ..gcore.utils import getMenuTrace, relax_menu_text, relax_menu_trace

logger = logging.getLogger(__name__)
respath = pathlib.Path(config['respath'])

class FuncToPanel(object):
    def __init__(self, func, *args):
        self.func = func
        self.args = args
                        
    def __call__(self):
        self.func(*self.args)
        
class CheckMenu(QtWidgets.QMenu):
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        if not parent is None:
            parent.addMenu(self)
        self.action_checked_calls = []
        self.action_enable_calls = []
        
    def showEvent(self, event):
        for action, checked_call in self.action_checked_calls:
            action.setChecked(checked_call())
            
        for action, enable_call in self.action_enable_calls:
            action.setEnabled(enable_call())            
            
    def addAction(self, *args, **kwargs):
        action = super().addAction(*args, **kwargs)
        if action is None: action = args[0]
        if not kwargs.get('checkcall', None) is None:
            action.setCheckable(True)
            self.action_checked_calls.append((action, kwargs.get('checkcall')))
        if not kwargs.get('enablecall', None) is None:
            self.action_enable_calls.append((action, kwargs.get('enablecall')))            
        return action        

class PanelsMenu(QMenu):

    def __init__(self, parent, name, categories, func):
        super().__init__(name, parent)
        qapp = QApplication.instance() 
        self.panels = qapp.panels
        self.categories = categories
        self.panel = func.__self__
        self.func = func
        
    def showEvent(self, event):
        self.initactions()        

    def initactions(self):
        self.clear()
        
        self.actions = []
       
        for category in self.categories:
            if not category in self.panels.keys(): continue
            panels = self.panels[category]
            keys = sorted(panels.keys())
            for panid in keys:                                            
                if category == self.parent().category and panid == self.parent().panid: continue
                panel = panels[panid]
                action = QAction(panel.windowTitle())
                action.setCheckable(True)
                action.setChecked((category, panel.panid) in self.panel.bindings)
                action.triggered.connect(FuncToPanel(self.func, category, panel.panid))
                self.addAction(action)
                self.actions.append(action)  
                
            self.addSeparator()

class MyStatusBar(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent) 
        hboxlayout = QtWidgets.QHBoxLayout()
        hboxlayout.setContentsMargins(0, 0, 0, 0)               
        self.setLayout(hboxlayout)
        self.setMinimumWidth(100)             
        
        fontmetric = QtGui.QFontMetrics(self.font())
        fontheight = fontmetric.height()
        self.setFixedHeight(fontheight + 2)        

        pal = self.palette()  
        pal.setColor(QtGui.QPalette.Background, QtGui.QColor(192,192,192))     
        self.setPalette(pal)        
        self.setAutoFillBackground(True)
        
    def addWidget(self, widget, stretch=0, alignment=None):
        widget.setParent(self)
        if not alignment is None:
            self.layout().addWidget(widget, stretch, alignment) 
        else:
            self.layout().addWidget(widget, stretch) 

    def removeWidget(self, widget):
        widget.setParent(None)
        self.layout().removeWidget(widget)             
            
        
def thisPanel(widget):
    while not (isinstance(widget, BasePanel) or widget is None):    
        if hasattr(widget, 'parentWidget'):
            #Used by matplotlib widgets
            parent_attr = widget.parentWidget
        else:
            parent_attr = widget.parent            
        widget = parent_attr()
    return widget

def selectThisPanel(widget):
    thisPanel(widget).select()
       

class BasePanel(QMainWindow):
    panelCategory = None
    panelShortName = None
    userVisible = True
        
    def __init__(self, parent, panid=None, category='console'):
        super().__init__()        
        self.baseWindowName = None               
        
        self.category = category                
        #Also creates the category if not exists
        id_exists = self.qapp.panels.id_exists(self.category, panid)        
        if panid is None: panid = self.qapp.panels.new_id(self.category)
        self.panid = panid            
        self.qapp.panels[self.category][self.panid] = self                
        
        self.bindings = []
        
        self.long_title = self.short_title        
        self.setWindowTitle(self.short_title)
        
        self.setFocusPolicy(Qt.StrongFocus)
        #self.setAttribute(Qt.WA_DeleteOnClose, True)        
        
        self.use_global_menu = True
        #self.myMainWidget = MyMainWidget()
        #super().setCentralWidget(self.myMainWidget)
        
        selIcon = QIcon()
        selIcon.addFile(str(respath / 'icons' / 'mark_16px.png'), state=QIcon.On)
        selIcon.addFile(str(respath / 'icons' / 'unmark_16px.png'), state=QIcon.Off)        
        
        self.setAutoFillBackground(True)        
        
        self.statusBar().setSizeGripEnabled(False)        

    @property
    def qapp(self):
        return QApplication.instance() 

    @property
    def short_title(self):
        return f'{self.category}#{self.panid}'
    
    @property
    def baseWindow(self):
        #return self.qapp.windows.get(self.baseWindowName, None)        
        container = self.get_container()
        if not container is None:
            return container.parent()
        else:
            return None
            
    def duplicate(self, floating=False):             
        newpanel = gui.qapp.panels.new_panel(type(self), None, None, floating=floating)        
        return newpanel

    def addMenuItem(self, menu, text, triggered, checkcall=None, enabled=True, statusTip=None, icon=None, enablecall=None):
    
        menuTrace = getMenuTrace(menu)
        menuTrace.append(text)
        catMenyShortCuts = self.qapp.menuCallShortCuts.get(self.category, {})
        keySequence = catMenyShortCuts.get(relax_menu_trace(menuTrace), None)
        
        # if not keySequence is None:
            # Note that the \t will place the keySequence in a nice second column
            # Like a real shortcut
            # text = f'{text}\t[{keySequence}]'            
        
        action = QAction(text, self, enabled=enabled, statusTip=statusTip)
        #action = TriggerlessShortcutAction(text, self, enabled=enabled, statusTip=statusTip)
        action.triggered.connect(triggered)
        
        if not keySequence is None:
            action.setShortcut(QtGui.QKeySequence(keySequence))
            #Disable the shortcut,
            #it is handled by the qapp.panelsDialog window
            #Shortcuts are send to the correct selected panel
            action.setShortcutContext(Qt.WidgetShortcut)
        
        if isinstance(icon, str):
            icon = QtGui.QIcon(str(respath / 'icons' / 'px16' / icon))
        
        if not icon is None:
            action.setIcon(icon)
            
        menu.addAction(action, checkcall=checkcall, enablecall=enablecall)      
        return action
    
    def addBaseMenu(self, bindCategories=[]):        
        self.bindMenu = PanelsMenu(self, 'bind to', bindCategories, self.toggleBindingTo)        
        self.menuBar().hide()        

    def toggleBindingTo(self, category, panid):
        added = self.addBindingTo(category, panid)
        if added is None:
            self.removeBindingTo(category, panid)        

    def addBindingTo(self, category, panid):        
        try:
            index = self.bindings.index((category, panid))            
            return None
        except ValueError:
            pass
            
        targetPanel = self.qapp.panels[category][panid]
        self.bindings.append((category, panid))
        
        return targetPanel

    def removeBindingTo(self, category, panid):
        try:
            index = self.bindings.index((category, panid))
        except ValueError:
            logger.debug(f'{category, panid} not in to the bindings')
            return None
            
        targetPanel = self.qapp.panels[category][panid]
        self.bindings.pop(index)       
        
        return targetPanel

    def panIdsOfBounded(self, category):
        return [bindpanid for bindcat, bindpanid in self.bindings if bindcat == category]
        
    def bindedPanel(self, category, pos=0):
        panids = self.panIdsOfBounded(category)
        if len(panids) == 0:
            return None
        return self.qapp.panels[category][panids[pos]]        
        
    def select(self):
        thisWasSelected = self.isSelected()
        self.qapp.panels.move_to_end(self)               
        
        if self.category in self.qapp.panels.ezm.bindGroups.keys():
            bindButton = self.qapp.panels.ezm.bindGroups[self.category].button(self.panid)
            if not bindButton is None:
                bindButton.setChecked(True)        

        if not self.use_global_menu:
            self.menuBar().show()            
        else:
            self.push_menu_to_main()           

        baseWindow = self.baseWindow
        if not baseWindow is None:
            baseWindow.setPanelInfo(self)      

        return thisWasSelected
        
    def isSelected(self):
        return self.qapp.panels.selected(self.category).panid == self.panid

    def push_menu_to_main(self):
        mainWindow = self.baseWindow 
        if mainWindow is None: return
        mainWindow.remove_panel_menu()                     
        
        for child in self.menuBar().children():            
            if isinstance(child, QMenu) and child.title() != '':  
                mainWindow.menuBar().addMenu(child)
                        
        self.menuBar().hide()

    def pull_menu_from_main(self):
        mainWindow = self.baseWindow
        if mainWindow is None: return
        menubar = self.menuBar()        
        mainWindow.remove_panel_menu()
        menubar.show()            

    def toggleMenu(self):
        self.use_global_menu = not self.use_global_menu
        self.select()
        
    def toggleStatusBar(self):
        statusBar = self.statusBar()
        if statusBar.isVisible():
            statusBar.hide()
        else:
            statusBar.show()        

    def detach(self):
        statusBar = self.statusBar()            
        self.setParent(None)          
        
        if self.use_global_menu:
            self.pull_menu_from_main()                                  

    def show_me(self):        
        container = self.get_container()
        if not container is None:
            window = container.parent()
            if window.isMinimized():
                window.showNormal()
            else:
                window.show()
            window.raise_()
            gui.qapp.setActiveWindow(window)            
        else:
            self.show()

    def unregister(self):
        self.qapp.panels[self.category].pop(self.panid)

    @classmethod
    def userPanelClasses(cls):
        l = []
        for SubClass in cls.__subclasses__():
            l.extend(SubClass.userPanelClasses())
            l.append((SubClass.panelCategory, SubClass))
            
        if cls is BasePanel:
            result = dict()
            
            for category, Cls in l:
                if not category in result.keys():
                    result[category] = []
                result[category].append(Cls)
            return result
        else:
            return l      

    def get_container(self):                
        from ..ezdock.ezdock import DockContainer
        
        candidate = self        
        while not (candidate is None or isinstance(candidate, DockContainer)):
            candidate = candidate.parent()
        return candidate        
        
    def mousePressEvent(self, event):
        self.select()

    def close_panel(self):
        container = self.get_container()
        window, laystruct = container.detach('panel', self.category, self.panid, False)
        self.qapp.processEvents()
        self.qapp.panels.removeBindingsTo(self.category, self.panid)
        self.unregister()
        window.unregister()       
        self.deleteLater()
        window.deleteLater()



"""Named mutex handling (for Win32).

See README.md or https://github.com/benhoyt/namedmutex for a bit more
documentation.

This code is released under the new BSD 3-clause license:
http://opensource.org/licenses/BSD-3-Clause

"""

import ctypes
from ctypes import wintypes

# Create ctypes wrapper for Win32 functions we need, with correct argument/return types
_CreateMutex = ctypes.windll.kernel32.CreateMutexA
_CreateMutex.argtypes = [wintypes.LPCVOID, wintypes.BOOL, wintypes.LPCSTR]
_CreateMutex.restype = wintypes.HANDLE

_WaitForSingleObject = ctypes.windll.kernel32.WaitForSingleObject
_WaitForSingleObject.argtypes = [wintypes.HANDLE, wintypes.DWORD]
_WaitForSingleObject.restype = wintypes.DWORD

_ReleaseMutex = ctypes.windll.kernel32.ReleaseMutex
_ReleaseMutex.argtypes = [wintypes.HANDLE]
_ReleaseMutex.restype = wintypes.BOOL

_CloseHandle = ctypes.windll.kernel32.CloseHandle
_CloseHandle.argtypes = [wintypes.HANDLE]
_CloseHandle.restype = wintypes.BOOL

class NamedMutex(object):
    """A named, system-wide mutex that can be acquired and released."""

    def __init__(self, name, acquired=False):
        """Create named mutex with given name, also acquiring mutex if acquired is True.
        Mutex names are case sensitive, and a filename (with backslashes in it) is not a
        valid mutex name. Raises WindowsError on error.
        
        """
        self.name =  name if isinstance(name, bytes) else name.encode()
        self.acquired = acquired
        self._create_mutext()
        if acquired:
            self.acquire()
            
    def _create_mutext(self):
        ret = _CreateMutex(None, False, self.name)
        if not ret:
            raise ctypes.WinError()
        self.handle = ret        
            
    def __getstate__(self):
        state = dict()
        state['name'] = self.name
        #Should an acquired lock be pickable?
        #It would block the receiving side
        state['acquired'] = self.acquired
        return state
    
    def __setstate__(self, state):
        self.name = state['name']      
        self.acquired = state['acquired']
        self._create_mutext()        
        if self.acquired:
            self.acquire()        

    def acquire(self, timeout=None):
        """Acquire ownership of the mutex, returning True if acquired. If a timeout
        is specified, it will wait a maximum of timeout seconds to acquire the mutex,
        returning True if acquired, False on timeout. Raises WindowsError on error.
        
        """
        if timeout is None:
            # Wait forever (INFINITE)
            timeout = 0xFFFFFFFF
        else:
            timeout = int(round(timeout * 1000))
        ret = _WaitForSingleObject(self.handle, timeout)
        if ret in (0, 0x80):
            # Note that this doesn't distinguish between normally acquired (0) and
            # acquired due to another owning process terminating without releasing (0x80)
            self.acquired = True
            return True
        elif ret == 0x102:
            # Timeout
            self.acquired = False
            return False
        else:
            # Waiting failed
            raise ctypes.WinError()

    def release(self):
        """Relase an acquired mutex. Raises WindowsError on error."""
        ret = _ReleaseMutex(self.handle)
        if not ret:
            raise ctypes.WinError()
        self.acquired = False

    def close(self):
        """Close the mutex and release the handle."""
        if self.handle is None:
            # Already closed
            return
        ret = _CloseHandle(self.handle)
        if not ret:
            raise ctypes.WinError()
        self.handle = None

    __del__ = close

    def __repr__(self):
        """Return the Python representation of this mutex."""
        return '{0}({1!r}, acquired={2})'.format(
                self.__class__.__name__, self.name, self.acquired)

    __str__ = __repr__

    # Make it a context manager so it can be used with the "with" statement
    def __enter__(self):
        self.acquire()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.release()
